/****************************************************************************
**
** Copyright (C) 2011 LP-Research
** All rights reserved.
** Contact: LP-Research (klaus@lp-research.com)
**
** This file is part of the Open Motion Analysis Toolkit (OpenMAT).
**
** OpenMAT is free software: you can redistribute it and/or modify it under 
** the terms of the GNU General Public License as published by the Free 
** Software Foundation, either version 3 of the License, or (at your option) 
** any later version.
** 
** OpenMAT is distributed in the hope that it will be useful, but WITHOUT 
** ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU \ General Public License 
** for more details.
** 
** You should have received a copy of the GNU General Public License along 
** with the OpenMAT library. If not, see <http://www.gnu.org/licenses/>.
**
****************************************************************************/

#include "AndroidBluetooth.h"

AndroidBluetooth::AndroidBluetooth(CalibrationData *configData, JavaVM *thisVm, jobject bluetoothAdapter) :
	LpmsIoInterface(configData),
	thisVm(thisVm),
	bluetoothAdapter(bluetoothAdapter)
{
	isOpen = false;
	
	JNIEnv* jniEnv;
	int status;

	status = thisVm->GetEnv((void **) &jniEnv, JNI_VERSION_1_4);
	if (status != JNI_OK) {
		LOGE("[AndroidBluetooth] Couldn't get environment in GUI thread");
		return;
	}	
	
	blueClass = jniEnv->FindClass("com/openmat/lpmsmonitor/BluetoothInterface");
	if (blueClass == NULL) {
		LOGE("[AndroidBluetooth] Couldn't find class BluetoothInterface");
		return;
	}
}

AndroidBluetooth::~AndroidBluetooth(void)
{
	close();
}
	
long long AndroidBluetooth::getConnectWait(void) 
{ 
	return 6000000; 
}

void AndroidBluetooth::listDevices(vector<DeviceListItem>* deviceList)
{
	// CALL JAVA
}

bool AndroidBluetooth::connect(string deviceId)
{
	// abm.lock();

	this->bluetoothAddress = deviceId;
	
	LOGE("[AndroidBluetooth] Starting connect");
	
	int status;	
	status = thisVm->AttachCurrentThread(&jniEnv, NULL);
	if (status != JNI_OK) {
		LOGE("[AndroidBluetooth] Couldn't attach to current thread");
		return false;
	}
	
	jmethodID blueConstructor = jniEnv->GetMethodID(blueClass, "<init>", "()V");
	if (blueConstructor == NULL) {
		LOGE("[AndroidBluetooth] Couldn't get constructor");
		return false;
	}	
		
	blueObject = jniEnv->NewObject(blueClass, blueConstructor);
	if (blueObject == NULL) {
		LOGE("[AndroidBluetooth] Couldn't create object");
		return false;
	}	
		
    connectMethod = jniEnv->GetMethodID(blueClass, "connect", "(Ljava/lang/String;Landroid/bluetooth/BluetoothAdapter;)V");
	if (connectMethod == NULL) {
		LOGE("[AndroidBluetooth] Couldn't get connect() function");
		return false;
	}		
		
	getReadBufferMethod = jniEnv->GetMethodID(blueClass, "readBuffer", "([B)I");
	if (getReadBufferMethod == NULL) {
		LOGE("[AndroidBluetooth] Couldn't get readBuffer() function");
		return false;
	}		
		
	/*getNumberBytesReadMethod = jniEnv->GetMethodID(blueClass, "numberBytesRead", "()I");
	if (getNumberBytesReadMethod == NULL) {
		LOGE("[AndroidBluetooth] Couldn't get numberBytesRead() function");
		return false;
	}*/			
	
    writeMethod = jniEnv->GetMethodID(blueClass, "writeBuffer", "([BI)V");
	if (writeMethod == NULL) {
		LOGE("[AndroidBluetooth] Couldn't get writeBuffer() function");
		return false;
	}		
		
	jbaOutputBuffer = jniEnv->NewByteArray(1024);	
	jbaInputBuffer = jniEnv->NewByteArray(1024);
	jbInputData = jniEnv->GetByteArrayElements((jbyteArray) jbaInputBuffer, 0);		
	jbOutputData = jniEnv->GetByteArrayElements((jbyteArray) jbaOutputBuffer, 0);	
	
	mm.reset();	
	oneTx.clear();
	rxState = PACKET_END;
	currentState = IDLE_STATE;
	waitForAck = false;
	waitForData = false;
	ackReceived = false;
	ackTimeout = 0;	
	lpmsStatus = 0;
	configReg = 0;
	dataReceived = false;
	pCount = 0;
	dataTimeout = 0;
	
	jstring jsAddress = jniEnv->NewStringUTF(deviceId.c_str());
    jniEnv->CallVoidMethod(blueObject, connectMethod, jsAddress, bluetoothAdapter);	
	
	isOpen = true;	
	
	boost::thread t(&AndroidBluetooth::readThread, this);
	t.detach();	
	
	// abm.unlock();	
	
	return true;
}

void AndroidBluetooth::close(void)
{
	if (isOpen == false) return;

	isOpen = false;
	
	// jniEnv->DeleteLocalRef(jbaOutputBuffer);	<<---!!!
	
	std::cout << "[LPMS-B] Closing connection" << std::endl;	
}	

void AndroidBluetooth::readThread(void)
{
	int status;	
	JNIEnv* readJniEnv;
	char rxBuffer[1024];
	
	status = thisVm->AttachCurrentThread(&readJniEnv, NULL);
	if (status != JNI_OK) {
		LOGE("[AndroidBluetooth] Couldn't attach to current thread");
		return;
	}

	while (isOpen == true) {
		jint numberBytesRead = readJniEnv->CallIntMethod(blueObject, getReadBufferMethod, jbaInputBuffer);	

		lpmsMutex.lock();
		for (unsigned int i=0; i<numberBytesRead; i++) {
			dataQueue.push((unsigned char) jbInputData[i]);
		}
		lpmsMutex.unlock();
		
		// LOGE("[AndroidBluetooth] Read: %d", numberBytesRead);
	}
}

bool AndroidBluetooth::read(char *rxBuffer, unsigned long *bytesReceived) 
{
	// LOGE("[AndroidBluetooth] Attempting read");

	if (isOpen == false) {
		LOGE("[AndroidBluetooth] Device not connected");	
		return false;
	}	
	
	float timeoutT;
	int n;
		
	jint numberBytesRead = jniEnv->CallIntMethod(blueObject, getReadBufferMethod, jbaInputBuffer);	
	
	for (int i=0; i<numberBytesRead; i++) {
		rxBuffer[i] = (char) jbInputData[i];
		//LOGE("[AndroidBluetooth] Byte %d: %x", i, rxBuffer[i]);
	}	

	*bytesReceived = numberBytesRead;
	
	// LOGE("[AndroidBluetooth] Called getReadBuffer()");	
	
	/*jsize jsiBufferSize = jniEnv->GetArrayLength((jbyteArray) joReadBuffer);
	int bufferSize = (int) jsiBufferSize;
	
	jbyte* jbData = jniEnv->GetByteArrayElements((jbyteArray) joReadBuffer, 0);

    jint jiNumberBytesRead = jniEnv->CallIntMethod(blueObject, getNumberBytesReadMethod);
	int numberBytesRead = (int) jiNumberBytesRead;
	// LOGE("[AndroidBluetooth] Got number of bytes read: %d", numberBytesRead);	
	
	// if (sizeof(rxBuffer) >= numberBytesRead && bufferSize >= numberBytesRead) {
		for (int i=0; i<numberBytesRead; i++) {
			rxBuffer[i] = (char) jbData[i];
			// LOGE("[AndroidBluetooth] Byte %d: %x", i, rxBuffer[i]);
		}
	
		*bytesReceived = numberBytesRead;
	// }
	
	jniEnv->ReleaseByteArrayElements((jbyteArray) joReadBuffer, jbData, 0);
	// jniEnv->DeleteLocalRef(joReadBuffer);
	
	// n = recv(sock, rxBuffer, sizeof(rxBuffer), 0);
	
	/* if (n == SOCKET_ERROR) {
		timeoutT = (float) mm.measure() / 1e6f;
		*bytesReceived = 0;
	} else {
		*bytesReceived = (unsigned long) n;
		
		mm.reset();
		timeoutT = 0;
	} */
	
	/*if (timeoutT > 3.0) {
		std::cout << "[LPMS-B] LPMS connection timeout has occured (device: " << bluetoothAddress.c_str() << ")." << std::endl;
				
		return false;
	}*/
	
	return true;
}

bool AndroidBluetooth::write(char *txBuffer, unsigned bufferLength)
{
	// LOGE("[AndroidBluetooth] Attempting write");
	
	if (isOpen == false) return false;

	lpmsMutex.lock();
	
	/* if (send(sock, txBuffer, bufferLength, 0) == SOCKET_ERROR) {
		return false;
	} */
	
	//jbaOutputBuffer = jniEnv->NewByteArray(bufferLength);		
	
	/*for (unsigned i=0; i<bufferLength; i++) {
		jbOutputData[i] = (jbyte) txBuffer[i];
		LOGE("[AndroidBluetooth] Write: %x", jbOutputData[i]);
	}*/		
	
	jniEnv->SetByteArrayRegion(jbaOutputBuffer, 0, bufferLength, (jbyte *)txBuffer);
	
	jniEnv->CallVoidMethod(blueObject, writeMethod, jbaOutputBuffer, (jchar) bufferLength);	
	
	lpmsMutex.unlock();
	
	return true;
}

bool AndroidBluetooth::sendModbusData(unsigned address, unsigned function, unsigned length, unsigned char *data)
{
	char txData[1024];
	unsigned int txLrcCheck;
	
	if (length > 1014) return false;

	txData[0] = 0x3a;
	txData[1] = address & 0xff;
	txData[2] = (address >> 8) & 0xff;
	txData[3] = function & 0xff;
	txData[4] = (function >> 8) & 0xff;
	txData[5] = length & 0xff;
	txData[6] = (length >> 8) & 0xff;
	
	for (unsigned int i=0; i < length; ++i) {
		txData[7+i] = data[i];
	}
	
	txLrcCheck = address;
	txLrcCheck += function;
	txLrcCheck += length;
	
	for (unsigned int i=0; i < length; i++) {
		txLrcCheck += data[i];
	}
	
	txData[7 + length] = txLrcCheck & 0xff;
	txData[8 + length] = (txLrcCheck >> 8) & 0xff;
	txData[9 + length] = 0x0d;
	txData[10 + length] = 0x0a;
	
	if (write(txData, length+11) == true) {
		return true;
	}
	
	return false;
}

bool AndroidBluetooth::parseModbusByte(unsigned char b)
{
	switch (rxState) {
	case PACKET_END:
		if (b == 0x3a) {
			rxState = PACKET_ADDRESS0;
			oneTx.clear();
		}
		break;
		
	case PACKET_ADDRESS0:
		currentAddress = b;
		rxState = PACKET_ADDRESS1;
		break;

	case PACKET_ADDRESS1:
		currentAddress = currentAddress + ((unsigned) b * 256);
		rxState = PACKET_FUNCTION0;
		break;

	case PACKET_FUNCTION0:
		currentFunction = b;
		rxState = PACKET_FUNCTION1;				
		break;

	case PACKET_FUNCTION1:
		currentFunction = currentFunction + ((unsigned) b * 256);
		rxState = PACKET_LENGTH0;
	break;

	case PACKET_LENGTH0:
		currentLength = b;
		rxState = PACKET_LENGTH1;
	break;
			
	case PACKET_LENGTH1:
		currentLength = currentLength + ((unsigned) b * 256);
		rxState = PACKET_RAW_DATA;
		rawDataIndex = currentLength;
	break;
			
	case PACKET_RAW_DATA:
		if (rawDataIndex == 0) {
			lrcCheck = currentAddress + currentFunction + currentLength;
			for (unsigned i=0; i<oneTx.size(); i++) lrcCheck += oneTx[i];
			lrcReceived = b;
			rxState = PACKET_LRC_CHECK1;
		} else {
			oneTx.push_back(b);
			--rawDataIndex;
		}
		break;
		
	case PACKET_LRC_CHECK1:
		lrcReceived = lrcReceived + ((unsigned) b * 256);			
		if (lrcReceived == lrcCheck) {
			parseFunction();
		} else {
			// std::cout << "[LPMS-B] Checksum fail in data packet" << std::endl;
		}
		
		rxState = PACKET_END;
		break;
	
	default:
		rxState = PACKET_END;		
		return false;
		break;
	}
	
	return true;
}

bool AndroidBluetooth::pollData(void) 
{	
	if (deviceStarted() == false) return false;
	
	return true;
}

bool AndroidBluetooth::deviceStarted(void)
{
	return isOpen;
}