/****************************************************************************
**
** Copyright (C) 2011 LP-Research
** All rights reserved.
** Contact: LP-Research (klaus@lp-research.com)
**
** This file is part of the Open Motion Analysis Toolkit (OpenMAT).
**
** OpenMAT is free software: you can redistribute it and/or modify it under 
** the terms of the GNU General Public License as published by the Free 
** Software Foundation, either version 3 of the License, or (at your option) 
** any later version.
** 
** OpenMAT is distributed in the hope that it will be useful, but WITHOUT 
** ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU \ General Public License 
** for more details.
** 
** You should have received a copy of the GNU General Public License along 
** with the OpenMAT library. If not, see <http://www.gnu.org/licenses/>.
**
****************************************************************************/

#include <string.h>
#include <jni.h>

#include <iostream>

#include "SensorManager.h"

// JNI finction declaration.
extern "C" {
	void Java_com_openmat_lpmsmonitor_LpmsMonitor_initSensorManagerJNI(JNIEnv* env, jobject obj, jobject blueAdapter);
	jint Java_com_openmat_lpmsmonitor_LpmsMonitor_addSensorJNI(JNIEnv* env, jobject obj, jstring address);
	void Java_com_openmat_lpmsmonitor_LpmsMonitor_getSensorData(JNIEnv* env, jobject obj, jfloatArray vertices);
	jstring Java_com_openmat_lpmsmonitor_LpSensor_stringFromJNI(JNIEnv* env, jobject obj);
	void Java_com_openmat_lpmsmonitor_LpmsMonitorRenderer_calcVertices(JNIEnv* env, jobject obj, jfloatArray v, jfloatArray q, jfloatArray outV);
	void Java_com_openmat_lpmsmonitor_LpmsSurfaceView_resetSensorOrientation(JNIEnv* env, jobject obj);
}

// Declares sensor manager, sensor as global objects.
SensorManager *sm;
JavaVM *thisVm;
LpSensor *lpmsDevice;
JNIEnv* jniEnv;

// Initializes sensor manager.
void Java_com_openmat_lpmsmonitor_LpmsMonitor_initSensorManagerJNI(JNIEnv* env, jobject obj, jobject blueAdapter)
{
	int status;

	status = thisVm->GetEnv((void **) &jniEnv, JNI_VERSION_1_4);
	if (status != JNI_OK) {
		LOGE("[LpSensorJNI] Couldn't get environment");
		return;
	}	

	// Creates new sensor manager. Passes java bluetooth adapter object.
	sm = new SensorManager(thisVm, blueAdapter);
}

// Adds one sensor to the sensor manager.
// TODO: Make lpmsDevice a vector of sensor handles to handle more than one sensor.
jint Java_com_openmat_lpmsmonitor_LpmsMonitor_addSensorJNI(JNIEnv* env, jobject obj, jstring address)
{
	std::string stdAddress(jniEnv->GetStringUTFChars((jstring) address, NULL));
	lpmsDevice = sm->addSensor(DEVICE_LPMS_B, stdAddress);
	
	return (jint) 1;
}

#include <Eigen/Dense>	

// Retrieves sensor data as orientation quaternion
void Java_com_openmat_lpmsmonitor_LpmsMonitor_getSensorData(JNIEnv* env, jobject obj, jfloatArray q)
{
	jboolean isCopy;	
	jfloat* qBuffer = jniEnv->GetFloatArrayElements(q, &isCopy);
	jsize length = jniEnv->GetArrayLength(q);
	
	lpmsDevice->getQuaternion(qBuffer);
	
	jniEnv->ReleaseFloatArrayElements(q, qBuffer, 0);
}

// Calculates vertices for cube demo.
// TODO: Move this to a different file as it is application specific code. 
void Java_com_openmat_lpmsmonitor_LpmsMonitorRenderer_calcVertices(JNIEnv* env, jobject obj, jfloatArray v, jfloatArray q, jfloatArray outV)
{
	LpVector4f lpQ;
	LpMatrix3x3f lpM;
	
	jboolean isCopy;	
	jfloat* vBuffer = jniEnv->GetFloatArrayElements(v, &isCopy);
	jsize length = jniEnv->GetArrayLength(v);	

	jfloat* qBuffer = jniEnv->GetFloatArrayElements(q, &isCopy);
	jfloat* outBuffer = jniEnv->GetFloatArrayElements(outV, &isCopy);	
	
	for (int i=0; i<4; i++) lpQ.data[i] = qBuffer[i];
	
	quaternionToMatrix(&lpQ, &lpM);
	
	Eigen::Matrix3f orientationM = Eigen::Matrix3f::Identity();
	orientationM << 	lpM.data[0][0], lpM.data[0][1], lpM.data[0][2],
						lpM.data[1][0], lpM.data[1][1], lpM.data[1][2],
						lpM.data[2][0], lpM.data[2][1], lpM.data[2][2];
						
	Eigen::Matrix3f scale;
	scale <<	1, 0, 0,
				0, 0, 1,
				0, -1, 0;
				
	Eigen::Matrix3f outputM;
	outputM = scale * orientationM;
	
	Eigen::Vector3f vertex;
	
	for (int i=0; i<length; i+=3) {
		vertex << vBuffer[i], vBuffer[i+1], vBuffer[i+2];
		
		vertex = outputM * vertex;
	
		outBuffer[i] = vertex(0);
		outBuffer[i+1] = vertex(1);
		outBuffer[i+2] = vertex(2);
	}
	
	jniEnv->ReleaseFloatArrayElements(q, qBuffer, 0);	
	jniEnv->ReleaseFloatArrayElements(v, vBuffer, 0);	
	jniEnv->ReleaseFloatArrayElements(outV, outBuffer, 0);	
}

// Resets sensor orientation offset.
void Java_com_openmat_lpmsmonitor_LpmsSurfaceView_resetSensorOrientation(JNIEnv* env, jobject obj)
{
	lpmsDevice->resetOrientation();
}	

// Returns JNI test string
// TODO: Remove
jstring Java_com_openmat_lpmsmonitor_LpmsMonitor_stringFromJNI(JNIEnv* env, jobject obj)
{
    return env->NewStringUTF("[LpSensorJNI] JNI test OK");
}

// Retrieves Java VM handle for Bluetooth access
jint JNI_OnLoad(JavaVM *vm, void *reserved) {
	thisVm = vm;
	JNIEnv* jniEnv;
	int status;
	
	LOGE("[LpSensorJNI] Retrieving JavaVM");

	status = thisVm->GetEnv((void **) &jniEnv, JNI_VERSION_1_4);
	if (status != JNI_OK) {
		LOGE("[LpSensorJNI] Couldn't retrieve environment handle");
		return false;
	}	
	
	return JNI_VERSION_1_4;
}